(ns awesome-app.core
  (:require [clojure.string :as string]
            [clojure.zip :as zip]
            [clojure.set :as s]
            #?(:cljs [cljs.reader])))

; Put all your super-awesome library code here. Re-frame will call
; `accept-input` with a string from the browser and the re-frame app database,
; and its return value will be fed verbatim back to ClojureScript/re-frame.

; Your app might return just a string, which will be packed into a <div> HTML
; tag
;(defn just-reverse [s]
;  (string/reverse s))

; Or it might return a Hiccup vector which re-frame/Reagent/React will render to
; HTML. In this case, <ol><li>...</ol>
;(defn enumerate [s]
;  (into [:ol]
;        (map (fn [element] [:li element])
;             (string/split s ""))))

; This is the gateway into your app. In this extremely trivial example,
; depending on how many inputs the user has submitted, the output will be either
; a reversed string or an enumerated list of characters. In this way, your
; application can know about the other commands the user has run.
;(defn accept-input [s & [db]]
;  (if (-> db :outputs count odd?)
;    (just-reverse s)
;    (enumerate s)))


;(defn now []
;  #?@(:cljs (.getTime (js/Date.)) 
;      ;:clj (java.lang.System/currentTimeMillis)
;            ))

(defn now []
  #?(,,,
       :cljs (.getTime (js/Date.)) 
       :clj (java.lang.System/currentTimeMillis)))

;(defn rs [s]
;  #?@(:cljs (cljs.reader/read-string s) 
;      :clj (read-string s)))

(defn rs [s]
  #?(,,,
       :cljs (cljs.reader/read-string s) 
       :clj (read-string s)))

(defn number-to-pair
  "Indexes into the enumeration of non-negative integer pairs using the 
  diagonal method often used to enumerate the rationals, except that everything
  is zero-based and we don't actually divide."
  [n]
  (loop [numerator 0 ;; terminology from algorithm for enumerating rationals
         denominator 0
         count 0]
    (if (= count n)
      [numerator denominator]
      (if (zero? numerator)
        (recur (inc denominator)
               0
               (inc count))
        (recur (dec numerator)
               (inc denominator)
               (inc count))))))
  
(defn indexed-term
  [i]
  (case i
    0 'x
    1 'y
    2 'z
    (let [[left right] (number-to-pair (- i 3))]
      (list '*
            (indexed-term left)
            (indexed-term right)))))

(defn next-term
  "Each call consumes one index and sets the value, returning the modified term source."
  [{:keys [base step indices]}]
  (if (empty? indices)
    (next-term {:base (+ base step)
                :step step
                :indices (shuffle (range base (+ base step)))
                :value nil})
    {:base base
     :step step
     :indices (rest indices)
     :value (indexed-term (first indices))}))

(defn term-source [step]
  (next-term {:base 0 :step step :indices [] :value nil}))

(defn num-vars [t]
  (count (filter #(some #{%} #{'x 'y 'z}) (flatten t))))

(defn codesize [c]
  (if (seq? c)
    (count (flatten c))
    1))

(defn at-index 
  "Returns a subtree of tree indexed by point-index in a depth first traversal."
  [tree point-index]
  (let [index (mod (Math/abs point-index) (codesize tree))
        zipper (zip/seq-zip tree)]
    (loop [z zipper i index]
      (if (zero? i)
        (zip/node z)
        (if (seq? (zip/node z)) 
          (recur (zip/next (zip/next z)) (dec i))
          (recur (zip/next z) (dec i)))))))

(defn insert-at-index
  "Returns a copy of tree with the subtree formerly indexed by
  point-index (in a depth-first traversal) replaced by new-subtree."
  [tree point-index new-subtree]
  (let [index (mod (Math/abs point-index) (codesize tree))
        zipper (zip/seq-zip tree)]
    (loop [z zipper i index]
      (if (zero? i)
        (zip/root (zip/replace z new-subtree))
        (if (seq? (zip/node z))
          (recur (zip/next (zip/next z)) (dec i))
          (recur (zip/next z) (dec i)))))))

(defn expand
  [t]
  (let [i (rand-int (codesize t))]
    (insert-at-index t i (cons '*
                               (shuffle [(at-index t i)
                                         (rand-nth '(x y z))])))))

(defn bud-term 
  "Returns a random term generated using the grow algorithm."
  [vars]
  (if (= vars 1)
    (rand-nth '(x y z))
    (loop [t (list '* (rand-nth '(x y z)) (rand-nth '(x y z)))]
      (if (>= (num-vars t) vars)
        t
        (recur (expand t))))))

(def a1 '((2 1 2)
          (1 0 0)
          (0 0 1)))

(def a2 '((2 0 2)
          (1 0 2)
          (1 2 1)))

(def a3 '((1 0 1)
          (1 2 0)
          (0 0 0)))

(def a4 '((1 0 1)
          (0 2 0)
          (0 1 0)))

(def a5 '((1 0 2)
          (1 2 0)
          (0 1 0)))

(def b1
  '((1 3 1 0)
    (3 2 0 1)
    (0 1 3 1)
    (1 0 2 0)))

(def b2
  '((2 1 0 3)
    (1 0 1 3)
    (1 0 3 0)
    (3 1 3 1)))

(def b3
  '((2 2 1 0)
    (0 2 3 3)
    (2 0 3 2)
    (0 0 3 0)))

(def p1
  '((3 2 0 4 0)
    (2 4 1 1 1)
    (1 1 3 0 1)
    (1 3 4 1 0)
    (0 1 0 1 0)))

(def p2
  '((2 1 2 4 1)
    (0 4 4 1 4)
    (0 4 4 1 3)
    (1 1 2 1 3)
    (1 1 0 0 3)))

(def i1
  '((1 3 0 4 1)
    (1 0 3 3 0)
    (3 4 3 0 1)
    (4 4 0 1 2)
    (0 3 1 4 4)))

(def i2
  '((2 3 2 2 1)
    (2 1 2 3 3)
    (1 3 4 3 3)
    (0 4 0 1 1)
    (3 2 4 1 1)))

(def i3
  '((1 3 4 3 3)
    (3 2 1 4 0)
    (3 1 0 2 2)
    (4 2 3 3 1)
    (2 2 4 1 1)))

(def i4
  '((0 3 2 1 3)
    (0 4 1 1 2)
    (3 4 3 2 0)
    (0 0 0 1 1)
    (0 4 1 3 1)))

(def d1
  '((1 4 0 1 3 0)
    (3 2 1 2 0 4)
    (2 0 3 5 1 3)
    (5 2 0 4 2 0)
    (2 1 3 0 5 1)
    (0 2 1 1 4 0)))

(def e1
  '((2 1 0 2 0 4 2)
    (4 5 0 3 3 3 0)
    (1 0 3 2 4 1 2)
    (0 4 3 1 2 5 1)
    (2 0 1 4 6 6 4)
    (0 6 2 4 3 4 2)
    (5 6 1 0 0 3 0)))

(declare params)
  
(defn merge-operations
  "Takes a left operation and a right operation and returns the operation of
  combining them with the given groupoid."
  [left-operation right-operation]
  (map #(nth (nth (:groupoid @params) %1) %2) left-operation right-operation))

(def term-operation
  ^{:doc "Returns a list of outputs of the term for each possible binding list in
    lexicographic order of xyz (with z varying the most quickly)."}
  (memoize 
    (fn [term]
      (let [sz (count (:groupoid @params))]
        (cond (number? term) (repeat (* sz sz sz) term)
              (symbol? term) (for [x (range sz)
                                   y (range sz)
                                   z (range sz)]
                               (case term x x y y z z))
              :else (merge-operations (term-operation (nth term 1))
                                      (term-operation (nth term 2))))))))

(defn discriminator-constraints
  "Returns a sequence of sets of acceptable answers for each binding list in
  lexicographic order of xyz (with z varying the most quickly), for a discriminator
  term of the given groupoid."
  [g]
  (let [n (count (first g))]
    (vec (for [x (range n) y (range n) z (range n)]
           #{(if (== x y) z x)}))))

(defn malcev-constraints
  "Returns a sequence of sets of acceptable answers for each binding list in
  lexicographic order of xyz (with z varying the most quickly), for a discriminator
  term of the given groupoid."
  [g]
  (let [n (count (first g))]
    (vec (for [x (range n) y (range n) z (range n)]
           (if (== x y) 
             #{z} 
             (if (== y z)
               #{x}
               (set (range n))))))))

(defn term-errors  
  "Returns a sequence of errors for each binding list in lexicographic order 
  of xyz (with z varying the most quickly), for the term relative to the 
  constraints. If the term produces a value consistent with the constraints
  for a binding list then the error will be 0, otherwise it will be 1."
  [term constraints]
  (map (fn [answer allowed]
         (if (some #{answer} allowed)
           0
           1))
       (term-operation term)
       constraints))

(defn side2-constraints
  "Returns a sequence of acceptable values (constraints) for each binding list in
  lexicographic order of xyz (with z varying the most quickly), relative to the
  specified overall (parent) constraints, for the second side of a term if the
  specified first side is filled with the specified term."
  [side1 side1-term parent-constraints]
  (let [g (:groupoid @params)
        error-sequences (let [sz (count g)]
                          (for [lz (range sz)]
                            (term-errors (case side1
                                           :left (list '* side1-term lz)
                                           :right (list '* lz side1-term))
                                         parent-constraints)))
        errors-for-assignments (apply map vector error-sequences)]
    (map (fn [errors]
           (set
             (map second
                  (filter #(zero? (first %))
                          (map vector errors (iterate inc 0))))))
         errors-for-assignments)))

(defn side1-constraints
  "Returns a sequence of acceptable values (constraints) for each binding list in
  lexicographic order of xyz (with z varying the most quickly), relative to the
  specified overall constraints, for the specified side as the first side to be filled
  in -- meaning that for each binding list some value for the other side will 
  produce correct results when combined with these 'acceptable' values."
  [side1 constraints]
  (let [g (:groupoid @params)
        constraints-for-constants (for [const (range (count g))]
                                    (side2-constraints (case side1 :left :right :right :left) 
                                                       const 
                                                       constraints))]
    (apply map s/union constraints-for-constants)))

(defn minimal-solution
  "If there is a minimal solution for the given constraints in (:minimals @params)
  then this will return a random one of them. Otherwise this returns nil."
  [constraints]
  (some #(if (zero? (reduce + (term-errors % constraints)))
           %
           false)
        (shuffle (:minimals @params))))

(defn drill
  "Searches for and returns a term that satisfies the given top-level-constraints."
  [top-level-constraints]
  (loop [stack (list {:constraints top-level-constraints})]
    (let [{:keys [constraints left right]} (first stack)]
      (if (and left right) ;; we have a complete term
        (if (= (count stack) 1) ;; and it's for the top-level problem
          (list '* left right)  ;; return it
          ;; else it's the solution to the parent problem
          (let [parent (second stack)
                solution-side (if (= (:left parent) :computing) :left :right)]
            (recur (cons (assoc parent solution-side (list '* left right))
                         (rest (rest stack))))))
        (if (not (or left right)) ;; new problem, nothing computed
          (if-let [ms (minimal-solution constraints)]
            (if (= (count stack) 1) ;; top level solution, return it
              ms
              (let [parent (second stack) ;; else build into parent and pop
                    solution-side (if (= (:left parent) :computing) :left :right)]
                (recur (cons (assoc parent solution-side ms)
                             (rest (rest stack))))))
            ;; no minimal solution, pick a side and compute it
            (let [computing-side (rand-nth [:left :right])]
              (recur (cons {:constraints (side1-constraints computing-side constraints)}
                           (cons {:constraints constraints computing-side :computing}
                                 (rest stack))))))
          ;; else one side is done, compute the remaining one
          (let [solved-side (if left :left :right)
                remaining-side (if left :right :left)
                solution (case solved-side :left left :right right)]
            (recur (cons {:constraints (side2-constraints solved-side solution constraints)}
                         (cons {:constraints constraints remaining-side :computing solved-side solution}
                               (rest stack))))))))))

(defn generate-minimals
  []
  (swap! params
         (fn [p]
           (assoc p
             :minimals 
             (let [num-minimals 
                   (reduce + 
                           (map #(int (Math/pow 3 %)) 
                                (map inc 
                                     (range (:max-length-of-minimals p)))))]
               (loop [index 0
                      minimals []]
                 (if (= (count minimals) num-minimals)
                   minimals
                   (recur (inc index)
                          (let [t (indexed-term index)]
                            (if (<= (count (filter #(some #{%} #{'x 'y 'z}) (flatten t)))
                                    (:max-length-of-minimals p))
                              (conj minimals t)
                              minimals))))))))))

(defn prefix->infix [t]
  (if (symbol? t)
    t
    (list (prefix->infix (nth t 1))
          '*
          (prefix->infix (nth t 2)))))

(def default-params
  {:groupoid a1
   :constraint-function discriminator-constraints
   :max-length-of-minimals 4
   :number-of-runs 1})

(def params (atom default-params))

(def output-buffer (atom ""))

(defn outputln [& args]
  (swap! output-buffer 
         str 
         (str (apply str (interleave args (repeat " ")))
              \newline)))


(defn termsearch
  "Conduct a term search, printing a report to standard output, using the 
  parameters in params except where overriden by provided arguments."
  [& args]
  (reset! output-buffer "")
  (outputln "Here are the vals of the input map: " (vals (rs (first args))))
  (outputln "=== termsearch started ===")
  ;(swap! params merge (apply hash-map args))
  (reset! params (merge default-params (apply hash-map args)))
  (generate-minimals)
  ;(outputln "parameters:" @params)
  (if (= (:constraint-function @params) :random-constraints)
    (do (outputln "=== beginning random constraints run ===")
      (let [random-terms (vec (repeatedly 10 #(bud-term 100)))
            term-ops (mapv term-operation random-terms)]
        (outputln "Random terms:" random-terms)
        (outputln "Term operations:" term-ops)
        (dotimes [run (count term-ops)]
          (outputln "=== starting run" run "===")
          (let [constraints (map hash-set (nth term-ops run))]
            (outputln "constraints:" constraints)
            (let [start-time (now)
                  t (drill constraints)
                  total-time (- (now) start-time)
                   ]
              ;(println "term:" t)
              ;(println "compact:" (apply str (remove #{\( \) \space} (str t))))
              (outputln "infix:" (apply str (remove #{\space} (str (prefix->infix t)))))
              (outputln "variables:" (count (filter #(not (= % \*))
                                                    (apply str (remove #{\( \) \space} (str t))))))
              (outputln "runtime (ms):" total-time)
              (outputln "=== run" run "completed ==="))))))
    (do
      (outputln "constraints:" ((:constraint-function @params) (:groupoid @params)))
      (dotimes [run (:number-of-runs @params)]
        (outputln "=== starting run" run "===")
        (let [start-time (now)
              t (drill ((:constraint-function @params) (:groupoid @params)))
              total-time (- (now) start-time)
               ]
          ;(println "term:" t)
          ;(println "compact:" (apply str (remove #{\( \) \space} (str t))))
          (outputln "infix:" (apply str (remove #{\space} (str (prefix->infix t)))))
          (outputln "variables:" (count (filter #(not (= % \*))
                                                (apply str (remove #{\( \) \space} (str t))))))
          (outputln "runtime (ms):" total-time)
          (outputln "=== run" run "completed ===")))))
  (outputln "=== all runs completed ===")
  @output-buffer
  )

(defn accept-input [s & [db]]
  (termsearch s)
  ;"this is a string"
  )